@@ -1,6 +1,6 @@
 
 /*
- * $Id: structs.h,v 1.276.2.3 1999/02/14 08:35:02 wessels Exp $
+ * $Id: structs.h,v 1.301.2.21 1999/10/22 19:34:06 wessels Exp $
  *
  *
  * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
@@ -75,6 +75,8 @@ struct _acl_proxy_auth_user {
     char *passwd;
     int passwd_ok;		/* 1 = passwd checked OK */
     long expiretime;
+    struct in_addr ipaddr;	/* IP addr this user authenticated from */
+    time_t ip_expiretime;
 };
 
 struct _acl_deny_info_list {
@@ -142,6 +144,7 @@ struct _aclCheck_t {
     struct in_addr src_addr;
     struct in_addr dst_addr;
     struct in_addr my_addr;
+    unsigned short my_port;
     request_t *request;
 #if USE_IDENT
     ConnStateData *conn;	/* hack for ident */
@@ -189,6 +192,11 @@ struct _relist {
     relist *next;
 };
 
+struct _sockaddr_in_list {
+    struct sockaddr_in s;
+    sockaddr_in_list *next;
+};
+
 #if DELAY_POOLS
 struct _delaySpec {
     int restore_bps;
@@ -231,6 +239,14 @@ struct _SquidConfig {
 	int pct;
 	size_t max;
     } quickAbort;
+#if HEAP_REPLACEMENT
+    char *replPolicy;
+#else
+    /* 
+     * Note: the non-LRU policies do not use referenceAge, but we cannot
+     * remove it until we find out how to implement #else for cf_parser.c
+     */
+#endif
     time_t referenceAge;
     time_t negativeTtl;
     time_t negativeDnsTtl;
@@ -240,19 +256,22 @@ struct _SquidConfig {
 	time_t read;
 	time_t lifetime;
 	time_t connect;
+	time_t peer_connect;
 	time_t request;
 	time_t pconn;
 	time_t siteSelect;
 	time_t deadPeer;
 	int icp_query;		/* msec */
+	int icp_query_max;	/* msec */
 	int mcast_icp_query;	/* msec */
 #if USE_IDENT
 	time_t ident;
 #endif
     } Timeout;
-    size_t maxRequestSize;
+    size_t maxRequestHeaderSize;
+    size_t maxRequestBodySize;
+    size_t maxReplyBodySize;
     struct {
-	ushortlist *http;
 	u_short icp;
 #if USE_HTCP
 	u_short htcp;
@@ -261,12 +280,21 @@ struct _SquidConfig {
 	u_short snmp;
 #endif
     } Port;
+    struct {
+	sockaddr_in_list *http;
+    } Sockaddr;
 #if SQUID_SNMP
     struct {
 	char *configFile;
 	char *agentInfo;
-	u_short localPort;
     } Snmp;
+#endif
+#if USE_WCCP
+    struct {
+	struct in_addr router;
+	struct in_addr incoming;
+	struct in_addr outgoing;
+    } Wccp;
 #endif
     char *as_whois_server;
     struct {
@@ -282,7 +310,7 @@ struct _SquidConfig {
     char *effectiveGroup;
     struct {
 	char *dnsserver;
-	char *redirect;
+	wordlist *redirect;
 	wordlist *authenticate;
 	char *pinger;
 	char *unlinkd;
@@ -291,6 +319,7 @@ struct _SquidConfig {
     int redirectChildren;
     int authenticateChildren;
     int authenticateTTL;
+    int authenticateIpTTL;
     struct {
 	char *host;
 	u_short port;
@@ -302,6 +331,7 @@ struct _SquidConfig {
     char *mimeTablePathname;
     char *visibleHostname;
     char *uniqueHostname;
+    wordlist *hostnameAliases;
     char *errHtmlText;
     struct {
 	char *host;
@@ -310,7 +340,6 @@ struct _SquidConfig {
 	u_short port;
     } Announce;
     struct {
-	struct in_addr tcp_incoming;
 	struct in_addr tcp_outgoing;
 	struct in_addr udp_incoming;
 	struct in_addr udp_outgoing;
@@ -373,9 +402,13 @@ struct _SquidConfig {
 #endif
 	int offline;
 	int redir_rewrites_host;
-	int persistent_client_posts;
 	int prefer_direct;
 	int strip_query_terms;
+	int redirector_bypass;
+	int ignore_unknown_nameservers;
+#if USE_CACHE_DIGESTS
+	int digest_generation;
+#endif
     } onoff;
     acl *aclList;
     struct {
@@ -393,6 +426,7 @@ struct _SquidConfig {
 #if USE_IDENT
 	acl_access *identLookup;
 #endif
+	acl_access *redirector;
     } accessList;
     acl_deny_info_list *denyInfoList;
     char *proxyAuthRealm;
@@ -424,8 +458,10 @@ struct _SquidConfig {
 #endif
     struct {
 	int icp_average;
+	int dns_average;
 	int http_average;
 	int icp_min_poll;
+	int dns_min_poll;
 	int http_min_poll;
     } comm_incoming;
     int max_open_disk_fds;
@@ -434,11 +470,22 @@ struct _SquidConfig {
 #if MULTICAST_MISS_STREAM
     struct {
 	struct in_addr addr;
+	int ttl;
 	unsigned short port;
 	char *encode_key;
     } mcast_miss;
 #endif
     HttpHeaderMask anonymize_headers;
+    char *coredump_dir;
+#if USE_CACHE_DIGESTS
+    struct {
+	int bits_per_entry;
+	int rebuild_period;
+	int rewrite_period;
+	int swapout_chunk_size;
+	int rebuild_chunk_percentage;
+    } digest;
+#endif
 };
 
 struct _SquidConfig2 {
@@ -517,9 +564,6 @@ struct _fde {
 	unsigned int nonblocking:1;
 	unsigned int ipc:1;
 	unsigned int called_connect:1;
-#ifdef OPTIMISTIC_IO
-	unsigned int calling_io_handler:1;
-#endif
     } flags;
     int bytes_read;
     int bytes_written;
@@ -601,12 +645,13 @@ struct _HttpHdrExtField {
 struct _HttpHdrCc {
     int mask;
     int max_age;
+    int s_maxage;
 };
 
 /* http byte-range-spec */
 struct _HttpHdrRangeSpec {
-    size_t offset;
-    size_t length;
+    ssize_t offset;
+    ssize_t length;
 };
 
 /* There may be more than one byte range specified in the request.
@@ -620,7 +665,7 @@ struct _HttpHdrRange {
 /* http content-range header field */
 struct _HttpHdrContRange {
     HttpHdrRangeSpec spec;
-    size_t elength;		/* entity length, not content length */
+    ssize_t elength;		/* entity length, not content length */
 };
 
 /* some fields can hold either time or etag specs (e.g. If-Range) */
@@ -634,8 +679,8 @@ struct _TimeOrTag {
 struct _HttpHdrRangeIter {
     HttpHdrRangePos pos;
     const HttpHdrRangeSpec *spec;	/* current spec at pos */
-    size_t debt_size;		/* bytes left to send from the current spec */
-    size_t prefix_size;		/* the size of the incoming HTTP msg prefix */
+    ssize_t debt_size;		/* bytes left to send from the current spec */
+    ssize_t prefix_size;	/* the size of the incoming HTTP msg prefix */
     String boundary;		/* boundary for multipart responses */
 };
 
@@ -1043,6 +1088,7 @@ struct _peer {
     } mcast;
 #if USE_CACHE_DIGESTS
     PeerDigest *digest;
+    char *digest_url;
 #endif
     int tcp_up;			/* 0 if a connect() fails */
     time_t last_fail_time;
@@ -1059,6 +1105,7 @@ struct _peer {
     } carp;
 #endif
     char *login;		/* Proxy authorization */
+    time_t connect_timeout;
 };
 
 struct _net_db_name {
@@ -1182,7 +1229,7 @@ struct _store_client {
     STCB *callback;
     void *callback_data;
     StoreEntry *entry;		/* ptr to the parent StoreEntry, argh! */
-    int swapin_fd;
+    storeIOState *swapin_sio;
     struct {
 	unsigned int disk_io_pending:1;
 	unsigned int store_copying:1;
@@ -1206,9 +1253,7 @@ struct _MemObject {
     int nclients;
     struct {
 	off_t queue_offset;	/* relative to in-mem data */
-	off_t done_offset;	/* relative to swap file with meta headers! */
-	int fd;
-	void *ctrl;
+	storeIOState *sio;
     } swapout;
     HttpReply *reply;
     request_t *request;
@@ -1221,7 +1266,14 @@ struct _MemObject {
 	void *data;
     } abort;
     char *log_url;
+#if HEAP_REPLACEMENT
+    /* 
+     * A MemObject knows where it is in the in-memory heap.
+     */
+    heap_node *node;
+#else
     dlink_node lru;
+#endif
     int id;
     ssize_t object_sz;
     size_t swap_hdr_sz;
@@ -1239,8 +1291,12 @@ struct _StoreEntry {
     size_t swap_file_sz;
     u_short refcount;
     u_short flags;
-    int swap_file_number;
+    sfileno swap_file_number;
+#if HEAP_REPLACEMENT
+    heap_node *node;
+#else
     dlink_node lru;
+#endif
     u_short lock_count;		/* Assume < 65536! */
     mem_status_t mem_status:3;
     ping_status_t ping_status:3;
@@ -1249,18 +1305,43 @@ struct _StoreEntry {
 };
 
 struct _SwapDir {
-    char *path;
-    int l1;
-    int l2;
+    swapdir_t type;
+    fileMap *map;
     int cur_size;
     int max_size;
+    char *path;
+    int index;			/* This entry's index into the swapDirs array */
     int suggest;
-    fileMap *map;
-    int swaplog_fd;
     struct {
 	unsigned int selected:1;
 	unsigned int read_only:1;
     } flags;
+    STINIT *init;
+    STNEWFS *newfs;
+    struct {
+	STOBJOPEN *open;
+	STOBJCLOSE *close;
+	STOBJREAD *read;
+	STOBJWRITE *write;
+	STOBJUNLINK *unlink;
+    } obj;
+    struct {
+	STLOGOPEN *open;
+	STLOGCLOSE *close;
+	STLOGWRITE *write;
+	struct {
+	    STLOGCLEANOPEN *open;
+	    STLOGCLEANWRITE *write;
+	    void *state;
+	} clean;
+    } log;
+    union {
+	struct {
+	    int l1;
+	    int l2;
+	    int swaplog_fd;
+	} ufs;
+    } u;
 };
 
 struct _request_flags {
@@ -1284,6 +1365,49 @@ struct _request_flags {
     unsigned int internal:1;
 };
 
+struct _link_list {
+    void *ptr;
+    struct _link_list *next;
+};
+
+struct _storeIOState {
+    sfileno swap_file_number;
+    mode_t mode;
+    size_t st_size;		/* do stat(2) after read open */
+    off_t offset;		/* current offset pointer */
+    STIOCB *callback;
+    void *callback_data;
+    struct {
+	STRCB *callback;
+	void *callback_data;
+    } read;
+    struct {
+	unsigned int closing:1;	/* debugging aid */
+    } flags;
+    union {
+	struct {
+	    int fd;
+	    struct {
+		unsigned int close_request:1;
+		unsigned int reading:1;
+		unsigned int writing:1;
+	    } flags;
+	} ufs;
+	struct {
+	    int fd;
+	    struct {
+		unsigned int close_request:1;
+		unsigned int reading:1;
+		unsigned int writing:1;
+		unsigned int opening:1;
+	    } flags;
+	    const char *read_buf;
+	    link_list *pending_writes;
+	    link_list *pending_reads;
+	} aufs;
+    } type;
+};
+
 struct _request_t {
     method_t method;
     protocol_t protocol;
@@ -1301,14 +1425,18 @@ struct _request_t {
     time_t ims;
     int imslen;
     int max_forwards;
+    /* these in_addr's could probably be sockaddr_in's */
     struct in_addr client_addr;
     struct in_addr my_addr;
+    unsigned short my_port;
     HttpHeader header;
     char *body;
     size_t body_sz;
+    int content_length;
     HierarchyLogEntry hier;
     err_type err_type;
     char *peer_login;		/* Configured peer login:password */
+    time_t lastmod;		/* Used on refreshes */
 };
 
 struct _cachemgr_passwd {
@@ -1363,11 +1491,11 @@ struct _ErrorState {
 	unsigned int flag_cbdata:1;
     } flags;
     struct {
+	wordlist *server_msg;
 	char *request;
 	char *reply;
     } ftp;
     char *request_hdrs;
-    wordlist *ftp_server_msg;
 };
 
 /*
@@ -1461,6 +1589,7 @@ struct _StatCounters {
     double cputime;
     struct timeval timestamp;
     StatHist comm_icp_incoming;
+    StatHist comm_dns_incoming;
     StatHist comm_http_incoming;
     StatHist select_fds_hist;
     struct {
@@ -1518,7 +1647,7 @@ struct _tlv {
 
 struct _storeSwapLogData {
     char op;
-    int swap_file_number;
+    sfileno swap_file_number;
     time_t timestamp;
     time_t lastref;
     time_t expires;
@@ -1531,8 +1660,8 @@ struct _storeSwapLogData {
 
 /* object to track per-action memory usage (e.g. #idle objects) */
 struct _MemMeter {
-    size_t level;		/* current level (count or volume) */
-    size_t hwater_level;	/* high water mark */
+    ssize_t level;		/* current level (count or volume) */
+    ssize_t hwater_level;	/* high water mark */
     time_t hwater_stamp;	/* timestamp of last high water mark change */
 };
 
@@ -1570,6 +1699,7 @@ struct _ClientInfo {
 	int n_req;
 	int n_denied;
     } cutoff;
+    int n_established;		/* number of current established connections */
 };
 
 struct _CacheDigest {
@@ -1599,6 +1729,7 @@ struct _FwdState {
     int n_tries;
     struct {
 	unsigned int dont_retry:1;
+	unsigned int ftp_pasv_failed:1;
     } flags;
 };
 
@@ -1672,3 +1803,16 @@ struct _helper_server {
 struct _generic_cbdata {
     void *data;
 };
+
+struct _store_rebuild_data {
+    int objcount;		/* # objects successfully reloaded */
+    int expcount;		/* # objects expired */
+    int scancount;		/* # entries scanned or read from state file */
+    int clashcount;		/* # swapfile clashes avoided */
+    int dupcount;		/* # duplicates purged */
+    int cancelcount;		/* # SWAP_LOG_DEL objects purged */
+    int invalid;		/* # bad lines */
+    int badflags;		/* # bad e->flags */
+    int bad_log_op;
+    int zero_object_sz;
+};
